"""Summary."""
from __future__ import annotations

from typing import List, Optional

from ergo.key import Key


class Topic:
    """Summary."""

    def __init__(self, topic_str: Optional[str]):
        """Summary.

        Args:
            topic_str (str): Description
        """
        self._keys: List[Key] = []
        if topic_str:
            self._keys = [Key(key_str) for key_str in topic_str.split('.')]

    def overlap(self, other: Topic):
        return set(self._keys) & set(other._keys)

    def __str__(self) -> str:
        """Summary.

        Returns:
            str: Description
        """
        ret = '#'
        if self._keys:
            ret = '#.%s.#' % '.#.'.join(sorted([str(key) for key in self._keys]))
        return ret


class SubTopic(Topic):
    """Summary."""


class PubTopic(Topic):
    """Summary."""

    def __str__(self) -> str:
        """Summary.

        Returns:
            str: Description
        """
        ret = '.'.join(sorted([str(key) for key in self._keys]))
        return ret



"""Summary."""
import datetime
import os

import yaml
from colors import color

from ergo.amqp_invoker import AmqpInvoker
from ergo.config import Config
from ergo.flask_http_invoker import FlaskHttpInvoker
from ergo.function_invocable import FunctionInvocable
from ergo.http_gateway import HttpGatewayServer
from ergo.http_invoker import HttpInvoker
from ergo.schematic import graph as ergograph
from ergo.version import get_version


def format_date(sec: float) -> str:
    """Summary.

    Args:
        sec (float): Description

    Returns:
        str: Description

    """
    dtf: str = '%b %d %Y, %H:%M:%S.%f'
    return datetime.datetime.fromtimestamp(sec).strftime(dtf)[:-3]


def get_version_path() -> str:
    """Summary.

    Returns:
        str: Description

    """
    return os.path.dirname(os.path.abspath(__file__)) + '/version.py'


def load_config(*config_paths: str) -> Config:
    merged_data = {}
    paths = list(config_paths)
    for path in paths:
        with open(path, "r") as fh:
            file_data = yaml.safe_load(fh)
        if "namespace" in file_data:
            paths.append(file_data.pop("namespace"))
        merged_data.update(file_data)
    return Config(merged_data)


class ErgoCli:
    """Summary."""

    @property
    def prompt(self) -> str:
        """Summary.

        Returns:
            str: Description

        """
        return f'{color("ergo", fg="#33ff33")} {color("∴", fg="#33ff33")} '
        # return 'ergo ∴ '

    @property
    def intro(self) -> str:
        """Summary.

        Returns:
            str: Description

        """
        version: str = get_version()
        timestamp: str = format_date(os.path.getmtime(get_version_path()))
        intro: str = f'ergo {version} ({timestamp})\nType help or ? to list commands.'

        return str(color(intro, fg='#ffffff'))

    def run(self, config: Config, *args: str) -> int:
        """Summary.

        Args:
            config (str): Description
            *args (str): Description

        Returns:
            int: Description

        Raises:
            err: Description

        """
        try:
            host: FunctionInvocable = FunctionInvocable(config)
            for result in host.invoke(dict(zip([str(i) for i in range(len(args))], args))):
                print(str(result))
        except Exception as err:
            print(f'*** {err}')
            raise err
        return 0

    def use(self, name: str) -> int:
        """Summary.

        Args:
            name (str): Description

        Returns:
            int: Description

        """
        return 0

    def init(self, name: str) -> int:
        """Summary.

        Args:
            name (str): Description

        Returns:
            int: Description

        """
        try:
            os.makedirs(f'.ergo/{name}')
        except FileExistsError:
            pass
        self.use(name)
        return 0

    @staticmethod
    def gateway(ref: str) -> int:
        config = load_config(ref)
        server = HttpGatewayServer(config)
        return server.run()

    def http(self, func: str, *args: str) -> int:
        """Summary.

        Args:
            func (str): Description
            *args (str): Description

        Returns:
            int: Description

        """
        config = Config({'func': func})
        return self._http(config)

    def _http(self, config: Config):
        host: HttpInvoker = FlaskHttpInvoker(FunctionInvocable(config))
        host.start()
        return 0

    def amqp(self, config: Config, *args: str) -> int:
        """Summary.

        Args:
            ref (str): Description
            *args (str): Description

        Returns:
            int: Description

        """
        host: AmqpInvoker = AmqpInvoker(FunctionInvocable(config))
        host.start()
        return 0

    def start(self, ref: str, *args: str) -> int:
        """Summary.

        Args:
            ref (str): Description
            *args (str): Description

        Returns:
            int: Description

        """
        config = load_config(ref, *args)
        if config.protocol == 'amqp':
            return self.amqp(config)
        if config.protocol == 'http':
            return self._http(config)
        raise ValueError(f'unexpected protocol: {config.protocol}')

    def graph(self, *args: str) -> int:
        """Summary.

        Args:
            ref (str): root folder for graph

        Returns:
            int: Description

        """
        ergograph(list(args))
        return 0



import asyncio
from typing import AsyncGenerator, Dict, Tuple

import aio_pika
import aiomisc
import hypercorn.asyncio
import hypercorn.config
from quart import Quart, request

from ergo.amqp_invoker import set_param
from ergo.config import Config
from ergo.message import Message, decode, decodes, encodes
from ergo.topic import PubTopic, SubTopic
from ergo.util import defer_termination, instance_id, uniqueid

EVENT_LOOP_THREADS = 10
MAX_CONCURRENT_RPCS = 10**4
RPC_TIMEOUT = 60 * 60  # seconds
PORT = 80


class HttpGatewayServer:
    def __init__(self, config: Config) -> None:
        self._config = config
        self._loop = aiomisc.new_event_loop(pool_size=EVENT_LOOP_THREADS)
        self._exchange, self._queue = self._loop.run_until_complete(self._setup_amqp(config))
        self._concurrent_rpcs_limiter = asyncio.Semaphore(MAX_CONCURRENT_RPCS)
        self._rpc_return_ready: Dict[str, asyncio.Condition] = {}
        self._rpc_return_values: Dict[str, Message] = {}

    def run(self) -> int:
        rpc_consumer_loop = self._loop.create_task(self._run_rpc_consumer())
        try:
            self._loop.run_until_complete(self._run_server())
        finally:
            rpc_consumer_loop.cancel()
        return 0

    async def _run_server(self):
        app = Quart(__name__)

        @app.route("/<path:path>", methods=["GET", "POST"])
        async def route(path: str):
            with defer_termination():
                async with self._concurrent_rpcs_limiter:
                    return await route_inner(path)

        async def route_inner(path: str):
            topic = path.replace("/", ".")
            # TODO what do we do if multiple results are being yielded by a generator?
            #  We're not injecting the handler, so we can't inspect for that.
            result_coro = self._rpc(topic, request.args).__anext__()
            message_out: Message = await asyncio.wait_for(result_coro, timeout=RPC_TIMEOUT)
            return encodes(message_out)

        hypercorn_config = hypercorn.config.Config()
        hypercorn_config.bind = [f"0.0.0.0:{PORT}"]

        await hypercorn.asyncio.serve(app, hypercorn_config)

    async def _rpc(self, topic: str, data: dict) -> AsyncGenerator[Message, None]:
        message = decode(**data)
        message.key = topic
        message.scope.reply_to = instance_id()
        correlation_id = message.scope.correlation_id = uniqueid()

        self._rpc_return_ready[correlation_id] = asyncio.Condition()
        try:
            amqp_message = aio_pika.Message(body=encodes(message).encode("utf-8"))
            routing_key = str(PubTopic(message.key))
            await self._exchange.publish(amqp_message, routing_key)
            async with self._rpc_return_ready[correlation_id]:
                await self._rpc_return_ready[correlation_id].wait()
            yield self._rpc_return_values.pop(correlation_id)
        finally:
            del self._rpc_return_ready[correlation_id]

    async def _run_rpc_consumer(self):
        async for amqp_message in self._queue:
            amqp_message.ack()
            ergo_message = decodes(amqp_message.body.decode("utf-8"))
            correlation_id = ergo_message.scope.correlation_id
            self._rpc_return_values[correlation_id] = ergo_message
            async with self._rpc_return_ready[correlation_id]:
                self._rpc_return_ready[correlation_id].notify()

    async def _setup_amqp(self, config: Config) -> Tuple[aio_pika.Exchange, aio_pika.Queue]:
        host = self._config.host
        heartbeat = self._config.heartbeat
        broker_url = set_param(host, "heartbeat", str(heartbeat)) if heartbeat else host
        connection: aio_pika.Connection = await aio_pika.connect_robust(broker_url)
        channel: aio_pika.Channel = await connection.channel()
        exchange: aio_pika.Exchange = await channel.declare_exchange(name=config.exchange, passive=True)
        queue: aio_pika.Queue = await channel.declare_queue(name=f"gateway:{instance_id()}", exclusive=True)
        await queue.bind(exchange=exchange, routing_key=str(SubTopic(instance_id())))
        return exchange, queue



"""Summary.

Attributes:
    ERGO_CLI (TYPE): Description

"""
from typing import Tuple

import click
from click_default_group import DefaultGroup  # https://pypi.org/project/click-default-group/

from ergo.config import Config
from ergo.ergo_cli import ErgoCli
from ergo.ergo_cmd import ErgoCmd

ERGO_CLI = ErgoCli()


@click.group(cls=DefaultGroup, default='shell', default_if_no_args=True)
def main() -> int:
    """Summary."""


@main.command()
def shell() -> int:
    """Summary.

    Returns:
        int: Description

    """
    ErgoCmd(ERGO_CLI).cmdloop()
    return 0


@main.command()
@click.argument('ref', type=click.STRING)
@click.argument('arg', nargs=-1)
def run(config: Config, arg: Tuple[str]) -> int:
    """Summary.

    Args:
        config (str): Description
        arg (Tuple[str]): Description

    Returns:
        int: Description

    """
    return ERGO_CLI.run(config, *list(arg))


@main.command()
@click.argument('name', type=click.STRING)
def init(name: str) -> int:
    """Summary.

    Args:
        name (str): Description

    Returns:
        int: Description

    """
    return ERGO_CLI.init(name)


@main.command()
@click.argument('name', type=click.STRING)
def use(name: str) -> int:
    """Summary.

    Args:
        name (str): Description

    Returns:
        int: Description

    """
    return ERGO_CLI.use(name)


@main.command()
@click.argument('ref', type=click.STRING)
def gateway(ref: str):
    return ERGO_CLI.gateway(ref)


@main.command()
@click.argument('func', type=click.STRING)
@click.argument('arg', nargs=-1)
def http(func: str, arg: Tuple[str]) -> int:
    """Summary.

    Args:
        func (str): Path to a function.
        arg (Tuple[str]): Description

    Returns:
        int: Description

    """
    return ERGO_CLI.http(func, *list(arg))


@main.command()
@click.argument('ref', type=click.STRING)
@click.argument('arg', nargs=-1)
def amqp(config: Config, arg: Tuple[str]) -> int:
    """Summary.

    Args:
        config (str): Description
        arg (Tuple[str]): Description

    Returns:
        int: Description

    """
    return ERGO_CLI.amqp(config, *list(arg))


@main.command()
@click.argument('ref', type=click.STRING)
@click.argument('arg', nargs=-1)
def start(ref: str, arg: Tuple[str]) -> int:
    """Summary.

    Args:
        ref (str): Description
        arg (Tuple[str]): Description

    Returns:
        int: Description

    """
    return ERGO_CLI.start(ref, *list(arg))


@main.command()
@click.argument('ref', type=click.STRING)
@click.argument('arg', nargs=-1)
def graph(ref: str, arg: Tuple[str]) -> int:
    """Summary.

    Args:
        ref (str): Description
        arg (Tuple[str]): Description

    Returns:
        int: Description

    """
    return ERGO_CLI.graph(ref, *list(arg))



"""Summary."""
import inspect
from typing import List

from flask import Flask, request  # , abort

from ergo.http_invoker import HttpInvoker
from ergo.message import Message, decode, encodes


class FlaskHttpInvoker(HttpInvoker):
    """Summary."""

    def start(self) -> int:
        """Summary.

        Returns:
            int: Description

        """
        app: Flask = Flask(__name__)

        @app.route(self.route, methods=['GET', 'POST'])
        def handler() -> str:  # type: ignore
            """Summary.

            Returns:
                str: Description

            """
            data_in: Message = decode(**request.args)
            data_out: List[Message] = list(self.invoke_handler(data_in))
            if not inspect.isgeneratorfunction(self._invocable.func):
                data_out = data_out[0]
            return encodes(data_out)

        app.run(host='0.0.0.0', port=self._port)
        return 0



"""Summary."""
import logging
import signal
import socket
import threading
from contextlib import contextmanager
from typing import Callable, Dict
from urllib.parse import urlparse

import kombu
import kombu.exceptions
import kombu.message
from kombu.pools import producers

from ergo.function_invocable import FunctionInvocable
from ergo.invoker import Invoker
from ergo.message import Message, decodes, encodes
from ergo.topic import PubTopic, SubTopic
from ergo.util import extract_from_stack, instance_id

logger = logging.getLogger(__name__)

PREFETCH_COUNT = 1
TERMINATION_GRACE_PERIOD = 60  # seconds
# rabbitmq's recommended default https://www.rabbitmq.com/heartbeats.html#heartbeats-timeout
DEFAULT_HEARTBEAT = 60  # seconds.


def set_param(host: str, param_key: str, param_val: str) -> str:
    """Overwrite a param in a host string w a new value."""
    uri, new_param = urlparse(host), f'{param_key}={param_val}'
    params = [p for p in uri.query.split('&') if param_key not in p] + [new_param]
    return uri._replace(query='&'.join(params)).geturl()


def make_error_output(err: Exception) -> Dict[str, str]:
    """Make a more digestible error output."""
    orig = err.__context__ or err
    err_output = {
        'type': type(orig).__name__,
        'message': str(orig),
    }
    filename, lineno, function_name = extract_from_stack(orig)
    if None not in (filename, lineno, function_name):
        err_output = {**err_output, 'file': filename, 'line': lineno, 'func': function_name}
    return err_output


class AmqpInvoker(Invoker):
    """Summary."""

    def __init__(self, invocable: FunctionInvocable) -> None:
        super().__init__(invocable)

        heartbeat = self._invocable.config.heartbeat or DEFAULT_HEARTBEAT
        self._connection = kombu.Connection(self._invocable.config.host, heartbeat=heartbeat)
        self._exchange = kombu.Exchange(name=self._invocable.config.exchange, type="topic", durable=True, auto_delete=False)

        component_queue_name = f"{self._invocable.config.func}".replace("/", ":")
        if component_queue_name.startswith(":"):
            component_queue_name = component_queue_name[1:]
        self._component_queue = kombu.Queue(name=component_queue_name, exchange=self._exchange, routing_key=str(SubTopic(self._invocable.config.subtopic)), durable=False)
        instance_queue_name = f"{component_queue_name}:{instance_id()}"
        self._instance_queue = kombu.Queue(name=instance_queue_name, exchange=self._exchange, routing_key=str(SubTopic(instance_id())), auto_delete=True)
        error_queue_name = f"{component_queue_name}:error"
        self._error_queue = kombu.Queue(name=error_queue_name, exchange=self._exchange, routing_key=str(SubTopic(error_queue_name)), durable=False)

        self._terminating = threading.Event()
        self._pending_invocations = threading.Semaphore()
        self._handler_lock = threading.Lock()

    def start(self) -> int:
        signal.signal(signal.SIGTERM, self._handle_sigterm)
        signal.signal(signal.SIGINT, self._handle_sigterm)
        with self._connection:
            conn = self._connection
            consumer: kombu.Consumer = conn.Consumer(queues=[self._component_queue, self._instance_queue], prefetch_count=PREFETCH_COUNT)
            consumer.register_callback(self._start_handle_message_thread)
            consumer.consume()
            while not self._terminating.is_set():
                try:
                    # wait up to 1s for the next message before sending a heartbeat
                    conn.drain_events(timeout=1)
                except socket.timeout:
                    conn.heartbeat_check()
                except conn.recoverable_connection_errors:
                    logger.warning("connection closed. reviving.")
                    conn = self._connection.clone()
                    conn.ensure_connection()
                    consumer.revive(conn.channel())
                    consumer.consume()
        return 0

    def _start_handle_message_thread(self, body: str, message: kombu.message.Message):
        # _handle_sigterm will wait for _handle_message to release this semaphore
        self._pending_invocations.acquire(blocking=False)
        threading.Thread(target=self._handle_message, args=(body, message.ack)).start()

    def _handle_message(self, body: str, ack: Callable):
        # there may be up to PREFETCH_COUNT _handle_message threads alive at a time, but we want them to execute
        # sequentially to guarantee that messages are acknowledged in the order they're received
        with self._handler_lock:
            ergo_message = decodes(body)
            try:
                self._handle_message_inner(ergo_message)
            finally:
                ack()
                self._pending_invocations.release()

    def _handle_message_inner(self, message_in: Message):
        try:
            for message_out in self.invoke_handler(message_in):
                routing_key = str(PubTopic(message_out.key))
                self._publish(message_out, routing_key)
        except Exception as err:  # pylint: disable=broad-except
            message_in.error = make_error_output(err)
            message_in.traceback = str(err)
            self._publish(message_in, self._error_queue.name)

    def _publish(self, ergo_message: Message, routing_key: str):
        amqp_message = encodes(ergo_message).encode("utf-8")
        with self._producer() as producer:
            producer.publish(
                amqp_message,
                content_encoding="binary",
                exchange=self._exchange,
                routing_key=routing_key,
                retry=True,
                declare=[self._instance_queue, self._error_queue],
            )

    @contextmanager
    def _producer(self) -> kombu.Producer:
        with producers[self._connection].acquire(block=True) as conn:
            yield conn

    def _handle_sigterm(self, signum, *_):
        self._terminating.set()
        self._pending_invocations.acquire(blocking=True, timeout=TERMINATION_GRACE_PERIOD)
        self._connection.close()
        signal.signal(signal.SIGTERM, 0)
        signal.raise_signal(signum)



"""Summary."""
import importlib.util
import inspect
import os
import re
import sys
import warnings
from importlib.abc import Loader
from importlib.machinery import ModuleSpec
from types import ModuleType
from typing import Callable, Generator, Match, Optional

import pydash

from ergo.config import Config
from ergo.context import Context, Envelope
from ergo.message import Message
from ergo.scope import Scope
from ergo.topic import Topic
from ergo.types import TYPE_RETURN
from ergo.util import instance_id, print_exc_plus

DATA_KEY = "data"
CONTEXT_KEY = "context"


class FunctionInvocable:
    """Summary."""

    def __init__(self, config: Config) -> None:
        """Summary.

        Args:
            reference (str): Description

        """
        self._func: Optional[Callable[..., TYPE_RETURN]] = None  # type: ignore
        self._params: dict = {}
        self._config: Config = config
        self.inject()

    @property
    def config(self) -> Config:
        """Summary.

        Returns:
            Config: Description

        """
        return self._config

    @property
    def func(self) -> Optional[Callable[..., TYPE_RETURN]]:  # type: ignore
        """Summary.

        Returns:
            Optional[Callable[..., TYPE_RETURN]]: Description

        """
        return self._func

    @func.setter
    def func(self, arg: Callable[..., TYPE_RETURN]) -> None:  # type: ignore
        """Summary.

        Args:
            arg (Callable[..., TYPE_RETURN]): Description

        """
        self._func = arg

    def invoke(self, message_in: Message) -> Generator[Message, None, None]:
        """Invoke injected function.

        If func is a generator, will exhaust generator, yielding each response.
        If an exception occurs will re-raise with a stack trace.
        Func responses will not be percolated if they return None.

        Args:
            message_in (ergo.message.Message): Contents will be passed to injected function as keyword args.

        Raises:
            Exception: caught exception re-raised with a stack trace.

        """
        if not self._func:
            raise Exception('Cannot execute injected function')
        try:
            ctx = Context(message=message_in, config=self.config)
            kwargs = self.assemble_arguments(message_in, ctx)
            results = self._func(**kwargs)
            if not inspect.isgenerator(results):
                results = [results]
            for data_out in results:
                envelope = None
                if isinstance(data_out, Envelope):
                    envelope = data_out
                    data_out = envelope.data
                scope = ctx._scope
                if Topic(f"{self.config.subtopic}.{instance_id()}").overlap(Topic(scope.reply_to)):
                    # The current scope was initiated in conjunction with a request that was addressed to this
                    # component or instance. We assume that by handling this message we've resolved
                    # the request, and may exit the current scope before proceeding. This frees handlers from
                    # needing to manually exit scope after receiving a request, or else publishing messages
                    # which will be routed back to them unto eternity.
                    assert scope.parent
                    scope = scope.parent
                if envelope and envelope.topic:
                    key = envelope.topic
                else:
                    key = self.config.pubtopic
                    if ctx.pubtopic != self.config.pubtopic:
                        key = ctx.pubtopic
                        warnings.warn("Context.pubtopic is going to be immutable in a future version of ergo. Use Context.envelope to override pubtopic.", category=DeprecationWarning)
                if envelope and envelope.reply_to:
                    scope = Scope(parent=scope)
                    scope.reply_to = envelope.reply_to
                elif scope.reply_to:
                    key = f"{key}.{scope.reply_to}"
                yield Message(data=data_out, scope=scope, key=key)

        except BaseException as err:
            raise Exception(print_exc_plus()) from err

    def assemble_arguments(self, message: Message, context: Context) -> dict:
        """
        assemble_arguments maps a message and a context to a dictionary of keyword arguments to pass to this invoker's
        handler.
        """
        kwargs = {}
        # this is the complete collection of data that this invoker's handler has access to
        exposed_data = {CONTEXT_KEY: context, DATA_KEY: message.data}
        for param, default in self._params.items():
            # ergo's canonical name for this param, which the configuration may have a custom mapping for
            ergo_param_name = self.config.args.get(param, param)
            argument = pydash.get(exposed_data, ergo_param_name) or pydash.get(exposed_data, f"{DATA_KEY}.{ergo_param_name}") or default
            # MissingArgument indicates that `param` is a positional parameter that we've failed to bind an argument
            # to, either because no argument was provided or because one was given the wrong name.
            if argument is not MissingArgument:
                kwargs[param] = argument
        return kwargs

    def inject(self) -> None:
        """Summary.

        Raises:
            Exception: Description

        """
        # [path/to/file/]<file>.<extension>[:[class.]method]]
        pattern: str = r'^(.*\/)?([^\.\/]+)\.([^\.]+):([^:]+\.)?([^:\.]+)$'  # (path/to/file/)(file).(extension):(method)
        matches: Optional[Match[str]] = re.match(pattern, self._config.func)
        if not matches:
            raise Exception(f'Unable to inject invalid referenced function {self._config.func}')

        path_to_source_file: str = matches.group(1)
        if not matches.group(1):
            path_to_source_file = os.getcwd()
        elif matches.group(1)[0] != '/':
            path_to_source_file = f'{os.getcwd()}/{matches.group(1)}'
        source_file_name: str = matches.group(2)
        source_file_extension: str = matches.group(3)
        sys.path.insert(0, path_to_source_file)

        spec: ModuleSpec = importlib.util.spec_from_file_location(source_file_name, f'{path_to_source_file}/{source_file_name}.{source_file_extension}')
        module: ModuleType = importlib.util.module_from_spec(spec)
        assert isinstance(spec.loader, Loader)  # see https://github.com/python/typeshed/issues/2793
        spec.loader.exec_module(module)

        scope: ModuleType = module
        if matches.group(4):
            class_name: str = matches.group(4)[:-1]
            scope = getattr(scope, class_name)

        method_name: str = matches.group(5)
        func = getattr(scope, method_name)

        if func:
            if inspect.ismethod(func.__call__):
                # func is a non-function class or instance, and we want to inject its __call__ method
                func = func.__call__
            self._func = func

            params = {}
            for name, info in inspect.signature(self._func).parameters.items():
                default = info.default
                if default is inspect.Parameter.empty:
                    # We use MissingArgument as a placeholder for positional parameters that don't have a default
                    # argument.
                    default = MissingArgument
                params[name] = default
            self._params = params


class MissingArgument:
    pass



"""Summary."""
import os
import sys
from typing import Dict, Generator, List, Tuple, Union

import graphviz
import yaml


def format_component(config: Dict[str, Union[None, str, List[str]]]) -> Tuple[str, str]:
    """Summary.

    Args:
        config (Dict[str, Union[None, str, List[str]]]): Description

    Returns:
        Tuple[str, str]: Description
    """
    return (f'component_{config.get("name")}', f'<<table border="0" cellborder="0"><tr><td bgcolor="#0071BD">{config.get("name")}</td></tr></table>>')


def format_topic(typestr: str, config: Dict[str, Union[None, str, List[str]]]) -> Generator[Tuple[str, str], None, None]:
    """Summary.

    Args:
        typestr (str): Description
        config (Dict[str, Union[None, str, List[str]]]): Description

    Yields:
        Generator[Tuple[str, str], None, None]: Description
    """
    tlist = config.get(typestr) or []
    if isinstance(tlist, str):
        tlist = [tlist]

    for topic_element in tlist:
        topic_str = '.'.join(sorted('&#x3a;'.join(topic_element.split(':')).split('.')))
        yield (f'topic_{topic_str}', topic_str)


def load_configs(folders: List[str]) -> List[Dict[str, Union[None, str, List[str]]]]:
    """Summary.

    Args:
        folders (List[str]): Description

    Returns:
        List[Dict[str, Union[None, str, List[str]]]]: Description
    """
    configs = []

    for folder in folders:
        for name in list_paths(folder):
            try:
                with open(f'{folder}/{name}/{name}.yaml', 'r', encoding='utf8') as stream:
                    config = yaml.safe_load(stream)
                    config['name'] = name
                    configs.append(config)
            except FileNotFoundError:
                pass
    return configs


def topics(dot: graphviz.Digraph, configs: List[Dict[str, Union[None, str, List[str]]]]) -> None:  # type: ignore[no-any-unimported]
    """Summary.

    Args:
        dot (graphviz.Digraph): Description
        configs (List[Dict[str, Union[None, str, List[str]]]]): Description
    """
    dot.attr('node', shape='none')

    for config in configs:
        for topic_element in format_topic('pubtopic', config):
            dot.edge(format_component(config)[0], topic_element[0])
            dot.node(*topic_element)
        for topic_element in format_topic('subtopic', config):
            dot.node(*topic_element)
            dot.edge(topic_element[0], format_component(config)[0])


def derived_topics(dot: graphviz.Digraph, configs: List[Dict[str, Union[None, str, List[str]]]]) -> None:  # type: ignore[no-any-unimported]
    """Summary.

    Args:
        dot (graphviz.Digraph): Description
        configs (List[Dict[str, Union[None, str, List[str]]]]): Description
    """
    for pub in configs:
        for sub in configs:
            for pub_topic in format_topic('pubtopic', pub):
                for sub_topic in format_topic('subtopic', sub):
                    if sub_topic[0] == pub_topic[0]:
                        continue

                    if all(topic_element in pub_topic[1].split('.') for topic_element in sub_topic[1].split('.')):
                        dot.edge(pub_topic[0], sub_topic[0])


def components(dot: graphviz.Digraph, configs: List[Dict[str, Union[None, str, List[str]]]]) -> None:  # type: ignore[no-any-unimported]
    """Summary.

    Args:
        dot (graphviz.Digraph): Description
        configs (List[Dict[str, Union[None, str, List[str]]]]): Description
    """
    dot.attr('node', shape='circle', width='1', color='#ffffff80', penwidth='2', fixedsize='true')

    for config in configs:
        dot.node(*format_component(config))


def graph(folders: List[str]) -> None:  # pylint: disable=too-many-branches
    """Summary.

    Args:
        folders (List[str]): Description
    """
    configs: List[Dict[str, Union[None, str, List[str]]]] = load_configs(folders)

    dot = graphviz.Digraph(comment='Component Diagram')
    dot.attr('graph', bgcolor='#0071BD', nodesep='1')
    dot.attr('edge', color='#ffffff80')
    dot.attr('node', fontcolor='#ffffff', fontname='arial')

    components(dot, configs)
    topics(dot, configs)
    derived_topics(dot, configs)

    dot.render('.ergo.gv', view=True)


def list_paths(path: str) -> List[str]:
    """Summary.

    Args:
        path (str): Description

    Returns:
        List[str]: Description
    """
    ret = []
    for file in os.listdir(path):
        directory = os.path.join(path, file)
        if os.path.isdir(directory):
            ret.append(file)
    return ret


if __name__ == '__main__':
    graph(sys.argv[1:])



"""Summary."""
import cmd  # https://docs.python.org/3/library/cmd.html
from typing import IO, Callable, List, Optional, cast

from ergo.ergo_cli import ErgoCli


class ErgoCmd(cmd.Cmd):
    """Summary."""

    intro: str = ''
    prompt: str = ''

    def __init__(self, cli: ErgoCli, completekey: str = 'tab', stdin: Optional[IO[str]] = None, stdout: Optional[IO[str]] = None) -> None:
        """Summary.

        Args:
            cli (ErgoCli): Description
            completekey (str, optional): Description
            stdin (Optional[IO[str]], optional): Description
            stdout (Optional[IO[str]], optional): Description

        """
        super().__init__(completekey, stdin, stdout)
        self._cli = cli
        ErgoCmd.intro = self._cli.intro
        ErgoCmd.prompt = self._cli.prompt

    def onecmd(self, line: str) -> bool:
        """Summary.

        Args:
            line (TYPE): Description

        Returns:
            TYPE: Description

        """
        splitline: List[str] = line.split()
        command: str = splitline[0]
        if hasattr(self, f'do_{command}') or not hasattr(self._cli, command):
            return super().onecmd(line)

        args: List[str] = splitline[1:]
        return cast(Callable[[str], bool], getattr(self._cli, command))(args[0], *args[1:])

    def do_exit(self, line: str) -> bool:
        """Summary.

        Args:
            line (str): Description

        Returns:
            int: Description

        """
        return True



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional

from ergo.util import uniqueid


@dataclass
class Scope:
    id: str = field(default_factory=uniqueid)
    metadata: dict = field(default_factory=dict)
    data: dict = field(default_factory=dict)
    parent: Optional[Scope] = None

    @property
    def reply_to(self) -> Optional[str]:
        return self.metadata.get("reply_to")

    @reply_to.setter
    def reply_to(self, value: str):
        self.metadata["reply_to"] = value

    @property
    def correlation_id(self) -> Optional[str]:
        return self.metadata.get("correlation_id")

    @correlation_id.setter
    def correlation_id(self, value: str):
        self.metadata["correlation_id"] = value



"""Summary."""
import dataclasses
import json
from dataclasses import dataclass, field
from typing import Any, Dict, Iterable, List, Optional, Union

import jsons

from ergo.scope import Scope


@dataclass
class Message:
    data: Any = field(default=None)
    key: Optional[str] = None
    log: List = field(default_factory=list)
    scope: Scope = field(default_factory=Scope)
    error: Optional[Dict[str, str]] = None
    traceback: Optional[str] = None


def decodes(s: str) -> Message:
    return decode(**jsons.loads(s))


def decode(**kwargs) -> Message:
    # if kwargs includes `data`, assume this message was sent by an upstream component, and the other kwargs are
    #   metadata
    # otherwise, assume this message came from outside of ergo, and bind all kwargs to `data`.
    if "data" not in kwargs:
        kwargs = {"data": kwargs or None}
    return jsons.load(kwargs, cls=Message)


def encodes(data: Union[Message, Iterable[Message]]) -> str:
    return json.dumps(data, cls=ErgoEncoder)


class ErgoEncoder(json.JSONEncoder):
    def default(self, o):
        if dataclasses.is_dataclass(o):
            return dataclasses.asdict(o)
        return super().default(o)



# unfortunately pylint is not able to recognize abstract subclasses
# pylint: disable=W0223

"""Summary."""
from ergo.function_invocable import FunctionInvocable
from ergo.invoker import Invoker


class HttpInvoker(Invoker):
    """Summary."""

    def __init__(self, invocable: FunctionInvocable) -> None:
        """Summary.

        Args:
            invocable (FunctionInvocable): Description

        """
        super().__init__(invocable)
        self._route: str = '/'
        self._port: int = 80

    @property
    def route(self) -> str:
        """Summary.

        Returns:
            str: Description

        """
        return self._route

    @route.setter
    def route(self, arg: str) -> None:
        """Summary.

        Args:
            arg (str): Description

        """
        self._route = arg

    @property
    def port(self) -> int:
        """Summary.

        Returns:
            int: Description

        """
        return self._port

    @port.setter
    def port(self, arg: int) -> None:
        """Summary.

        Args:
            arg (int): Description

        """
        self._port = arg



"""Summary."""
import copy
from collections import OrderedDict
from typing import Dict, Optional

from ergo.topic import PubTopic, SubTopic, Topic


class Config:
    """Summary."""

    # pylint: disable=too-many-instance-attributes
    # Eight is reasonable in this case.

    def __init__(self, config: Dict[str, str]):
        """Summary.

        Args:
            config (Dict[str, str]): Description
        """
        self._func: Optional[str] = config.get('func')
        self._namespace: Optional[str] = config.get('namespace', 'local')
        self._pubtopic: str = config.get('pubtopic')
        self._subtopic: str = config.get('subtopic')
        self._host: Optional[str] = config.get('host')
        self._exchange: Optional[str] = config.get('exchange')
        self._protocol: str = config.get('protocol', 'stack')  # http, amqp, stdio, stack
        self._heartbeat: Optional[str] = config.get('heartbeat')
        self._args: Optional[dict] = config.get('args')

    def copy(self):
        return copy.deepcopy(self)

    @property
    def args(self) -> dict:
        """Summary.

        Returns:
            TYPE: Description
        """
        return self._args or {}

    @args.setter
    def args(self, val: dict) -> None:
        """Summary.

        Returns:
            TYPE: Description
        """
        self._args = val

    @property
    def namespace(self) -> Optional[str]:
        """Summary.

        Returns:
            TYPE: Description
        """
        return self._namespace

    @property
    def subtopic(self) -> str:
        """Summary.

        Returns:
            TYPE: Description
        """
        return self._subtopic

    @property
    def pubtopic(self) -> str:
        """Summary.

        Returns:
            TYPE: Description
        """
        return self._pubtopic

    @pubtopic.setter
    def pubtopic(self, val: str) -> None:
        """Summary.

        Returns:
            TYPE: Description
        """
        self._pubtopic = val

    @property
    def func(self) -> Optional[str]:
        """Summary.

        Returns:
            TYPE: Description
        """
        return self._func

    @property
    def host(self) -> str:
        """Summary.

        Returns:
            TYPE: Description
        """
        return self._host or ''

    @property
    def exchange(self) -> str:
        """Summary.

        Returns:
            TYPE: Description
        """
        return self._exchange or 'primary'

    @property
    def protocol(self) -> str:
        """Summary.

        Returns:
            TYPE: Description
        """
        return self._protocol

    @property
    def heartbeat(self) -> Optional[int]:
        """Summary.

        Returns:
            TYPE: Description
        """
        return int(self._heartbeat) if self._heartbeat else None



"""Summary.

Attributes:
    VERSION (str): Description

"""
import subprocess
import sys

VERSION = '0.8.0-alpha'


def get_version() -> str:
    """Summary.

    Returns:
        str: Description

    """
    return VERSION


# wtc && match => ok
# wtc && incr => not ok
# changes && match => not ok
# changes && incr => ok
def main() -> None:
    """Summary."""
    ver: str = get_version()
    tag: str = subprocess.check_output(['git', 'describe', '--tags']).decode('utf-8')
    status: str = subprocess.check_output(['git', 'status']).decode('utf-8')
    try:
        if tag.index(ver) == 0 and 'working tree clean' not in status:  # if version hasn't changed
            print(f'Version {VERSION} must be incremented in src/version.py because codebase has changed')
            sys.exit(1)
        else:
            # print('not status')
            # sys.exit(0)
            print(get_version())
    except ValueError:
        print(get_version())


# def main() -> None:
#     """Summary."""
#     ver: str = get_version()
#     tag: str = subprocess.check_output(['git', 'describe', '--tags']).decode('utf-8')
#     status: str = subprocess.check_output(['git', 'status']).decode('utf-8')
#     clean = 'working tree clean' in status
#     match = False
#     try:
#         match = tag.index(ver) == 0
#     except ValueError:
#         pass

#     if clean and not match or match and not clean:
#         print('Version must be incremented with changes to codebase')
#         sys.exit(1)
#     else:
#         print(get_version())


if __name__ == '__main__':
    main()



"""Convenience Funcs for handling errors, logging, and monitoring."""
import re
import signal
import sys
import threading
import time
import traceback
from functools import lru_cache
from types import FrameType, TracebackType
from typing import List, Optional, Tuple
from uuid import uuid4

if sys.version_info >= (3, 8):
    from typing import TypedDict  # pylint: disable=C0412
else:
    from typing_extensions import TypedDict  # pragma: no cover


class LogStruct(TypedDict):
    """Summary."""

    ts: float
    mid: str
    cid: str


def log(rec: List[LogStruct]) -> List[LogStruct]:
    """Summary.

    Args:
        rec (List[Dict[str, Any]]): Description

    Returns:
        List[Dict[str, Any]]: Description
    """
    cid: str = uniqueid()
    if rec:
        cid = rec[-1].get('cid', cid)
    rec.append(LogStruct({'ts': timestamp(), 'mid': uniqueid(), 'cid': cid}))
    return rec


def uniqueid() -> str:
    """Generate unique id.

    Returns:
        str: unique hex id
    """
    return uuid4().hex


def timestamp() -> float:
    """Create timestamp for current time.

    Returns:
        int: current time
    """
    return time.time()


def get_stack() -> List[FrameType]:
    """Summary.

    Returns:
        List[FrameType]: Description
    """
    frm: Optional[FrameType] = None
    trcbk: Optional[TracebackType] = sys.exc_info()[2]
    while trcbk:
        if not trcbk.tb_next:
            frm = trcbk.tb_frame
            break
        trcbk = trcbk.tb_next  # pragma: no cover
    stack: List[FrameType] = []
    while frm:
        stack.append(frm)
        frm = frm.f_back
    return stack


def print_exc_plus() -> str:  # pragma: no cover
    """
    Summary.

    Print the usual traceback information, followed by a listing of all the
    local variables in each frame.

    Returns:
        str: Description
    """
    ret = ''
    stack = get_stack()
    stack.reverse()
    ret = f'{ret}\n{traceback.format_exc()}'
    ret = f'{ret}\nLocals by frame, innermost last'
    for frame in stack:
        ret = f'{ret}\nFrame {frame.f_code.co_name} in {frame.f_code.co_filename} at line {frame.f_lineno}'
        for key, value in frame.f_locals.items():
            ret = f'{ret}\n\t{key} = '
            # We have to be VERY careful not to cause a new error in our error
            # printer! Calling str(  ) on an unknown object could cause an
            # error we don't want, so we must use try/except to catch it --
            # we can't stop it from happening, but we can and should
            # stop it from propagating if it does happen!
            try:
                ret = f'{ret}\n{value}'
            except BaseException:  # pylint: disable=broad-except
                ret = f'{ret}\n<ERROR WHILE PRINTING VALUE>'

    return ret


def extract_from_stack(exc: BaseException) -> Tuple[Optional[str], Optional[str], Optional[str]]:
    """
    Extract minimal useful information from top of function stack.

    Returns:
        str: filename
        str: lineno
        str: function name
    """
    traceback_exception = traceback.TracebackException.from_exception(exc)
    stack_string = traceback_exception.stack.format()[-1]
    # File ".*", line n+, in s+\n
    prog = re.compile(r'File ".+/(\w+[.]py)", line (\d+), in (\w+)\n')
    match = prog.search(stack_string)
    if match:
        matches = match.groups()
        if len(matches) == 3:  # for mypy
            return matches[0], matches[1], matches[2]
    return None, None, None


_shutdown = threading.Event()
_termination_pending = threading.Event()


class defer_termination:
    """
    Use this context manager to temporarily postpone shutdown via SIGTERM.
    """

    def __enter__(self):
        if _termination_pending.is_set():
            _shutdown.wait()
        self._signum = None
        signal.signal(signal.SIGTERM, self._sigterm_handler)

    def __exit__(self, exc_type, exc_val, exc_tb):
        signal.signal(signal.SIGTERM, 0)
        if self._signum:
            signal.raise_signal(self._signum)

    def _sigterm_handler(self, signum, _):
        _termination_pending.set()
        self._signum = signum


@lru_cache(1)
def instance_id() -> str:
    return uniqueid()



from typing import Any, Optional

from ergo.config import Config
from ergo.message import Message
from ergo.scope import Scope
from ergo.util import instance_id


class Envelope:
    """
    Use this container to pass data to ergo with custom routing parameters.

    >>> ad_hoc_topic = "interested_party"
    >>> def my_handler(context: Context):
    ...     return context.envelope("my_return_val", topic=f"{context.pubtopic}.{ad_hoc_topic}")
    ...
    >>>
    """

    def __init__(self, data: Any, topic: Optional[str] = None, reply_to: Optional[str] = None):
        self.data = data
        self.topic = topic
        self.reply_to = reply_to


class Context:
    envelope = Envelope

    def __init__(self, message: Message, config: Config):
        self.pubtopic: str = config.pubtopic
        self._scope = message.scope

    @property
    def instance_id(self):
        return instance_id()

    def initiate_scope(self):
        self._scope = Scope(parent=self._scope)

    def exit_scope(self):
        self._scope = self._scope.parent

    def retrieve(self, key: str) -> Any:
        if self._scope:
            return self._scope.data[key]

    def store(self, key: str, value: Any):
        if self._scope:
            self._scope.data[key] = value
